# Copyright 2015  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import glob
import logging
import os
import re
import shutil
import StringIO
import subprocess
import sys
import tempfile
import time
import urllib
import urlparse

import cliapp

import icklib


class GitSpec(object):

    def __init__(self, url, branch, subdir):
        self.url = url
        self.branch = branch
        self.subdir = subdir


class Project(object):

    def __init__(self, name):
        self.name = name
        self.progress = None
        self.git_specs = []
        self.repo_signing_key = None
        self.commands = []
        self.local_commands = []
        self.env = {}
        self.dput_targets = []

        self.pipelines = {}

    def set_progress(self, progress):
        self.progress = progress

    def add_git_spec(self, git_spec):
        self.git_specs.append(git_spec)

    def set_repo_signing_key(self, key_id):
        self.repo_signing_key = key_id

    def add_shell_command(self, command):
        self.commands.append(command)

    def add_local_command(self, command):
        self.local_commands.append(command)

    def add_dput_target(self, dput_target):
        self.dput_targets.append(dput_target)

    def set_build_pipelines(self, pipelines):
        self.pipelines = pipelines

    def set_env(self, env):
        self.env = env

    def build(self, statedir, targets):
        for pipeline_name, pipeline_steps in self.pipelines.items():
            pipeline = icklib.BuildPipeline(pipeline_name)
            pipeline.set_build_steps(pipeline_steps)
            run_state = icklib.RunState()
            run_state.pipeline = pipeline
            run_state.progress = self.progress
            run_state.logger = icklib.Logger()
            run_state.log_catcher = StringIO.StringIO()
            run_state.logger.add_output_file(run_state.log_catcher)
            self.run_state = run_state

            self.progress['pipeline'] = pipeline_name
            run_state.logger.important(
                'Build log starts {timestamp}',
                timestamp=self.current_timestamp())
            run_state.logger.important(
                'Building project {project}',
                project=self.name)
            run_state.logger.important(
                'Executing pipeline {name}', name=pipeline_name)

            run_state.started = time.time()
            try:
                pipeline.build(self, statedir, targets, run_state)
            except:
                if run_state.build_info is not None:
                    run_state.build_info.status = 'FAILURE'
                    run_state.build_info.duration = (
                        time.time() - run_state.started)
                    run_state.build_info.save()
                run_state.logger.important('')
                run_state.logger.important('Build status: FAILURE')
                run_state.logger.important(
                    'Build log ends {timestamp}',
                    timestamp=self.current_timestamp())
                raise

            if run_state.build_info is not None:
                run_state.logger.important(
                    'Build status: {status}',
                    status=run_state.build_info.status)
            run_state.logger.important(
                'Build log ends {timestamp}',
                timestamp=self.current_timestamp())

    def current_timestamp(self):
        timezone = '{sign}{h}{m}'.format(
            sign='-' if time.timezone < 0 else '+',
            h=abs(time.timezone) / 3600,
            m=(abs(time.timezone) % 3600) / 60)
        return '{timestamp} {timezone}'.format(
            timestamp=time.strftime('%Y-%m-%d %H:%M:%S'),
            timezone=timezone)

    def run_locally(self, argv, displayed_cmd, cwd):
        self.report_running_locally(displayed_cmd)
        self.run_state.logger.important(
            'On local shell run {command}',
            command=displayed_cmd)

        whole_argv = self.create_argv_with_env() + argv

        line_logger = LineLogger(self.progress, self.run_state.logger)
        with self.run_state.logger:
            self.run_state.logger.chatty('+ %s' % ' '.join(whole_argv))
            cliapp.runcmd(
                whole_argv,
                stderr=subprocess.STDOUT,
                stdout_callback=line_logger.log_lines,
                cwd=cwd)
            line_logger.log_rest()

        self.report_finished_on_target()

    def create_argv_with_env(self):
        env_argv = ['env']
        for name, value in self.env.items():
            env_argv.append('%s=%s' % (name, value))
        return env_argv

    def report_running_locally(self, displayed_cmd):
        progress = {
            'target': 'local',
            'command': '\\n'.join(displayed_cmd.split('\n')),
        }
        for i in range(self.run_state.progress.max_output_lines):
            progress['line%d' % i] = ''
        for key, value in progress.items():
            self.run_state.progress[key] = value

    def run_argv_on_target(self, target, argv, displayed_cmd, remote_cwd):
        self.report_running_on_target(target, displayed_cmd)
        self.run_state.logger.important(
            'On {address} run {command}',
            address=target.address, command=displayed_cmd)

        env_argv = ['env']
        for name, value in self.env.items():
            env_argv.append('%s=%s' % (name, value))
        whole_argv = self.create_argv_with_env() + argv

        line_logger = LineLogger(self.progress, self.run_state.logger)
        with self.run_state.logger:
            self.run_state.logger.chatty('+ %s' % ' '.join(whole_argv))
            target.ssh_runcmd(
                whole_argv,
                remote_cwd=remote_cwd,
                tty=False,
                stderr=subprocess.STDOUT,
                stdout_callback=line_logger.log_lines)
            line_logger.log_rest()

        self.report_finished_on_target()

    def report_running_on_target(self, target, displayed_cmd):
        progress = {
            'target': target.name,
            'command': '\\n'.join(displayed_cmd.split('\n')),
        }
        for i in range(self.run_state.progress.max_output_lines):
            progress['line%d' % i] = ''
        for key, value in progress.items():
            self.run_state.progress[key] = value

    def report_finished_on_target(self):
        keys = ['target', 'command']
        for i in range(self.run_state.progress.max_output_lines):
            keys.append('line%d' % i)
        for key in keys:
            self.run_state.progress[key] = ''

    def copy_cleanly_to_target(self, target, remote_tmp):
        self.run_state.logger.chatty(
            'Copy helper script (./cleanly) to target')
        local_cleanly_path = os.path.join(
            os.path.dirname(icklib.__file__), 'cleanly')
        remote_cleanly_path = os.path.join(remote_tmp, 'cleanly')
        cliapp.runcmd(
            ['rsync', '-s', local_cleanly_path,
             '%s:%s' % (target.address, remote_cleanly_path)])
        return remote_cleanly_path

    def run_cleanly(self, target, remote_git_dir, cleanly_path, args):
        self.run_argv_on_target(
            target,
            ['python', cleanly_path] + args,
            'cleanly ' + args[-1],
            remote_git_dir)


class LoadProjectInfo(icklib.BuildStep):

    def build(self):
        self.run_state.progress['step'] = 'Loading project state'
        self.run_state.logger.important('Loading project state')
        self.run_state.project_info = icklib.ProjectInformation()

        filename = self.statedir.get_project_state_path(self.project.name)
        self.run_state.project_info.set_filename(filename)
        dirname = os.path.dirname(filename)

        if not os.path.exists(dirname):
            os.makedirs(dirname)
        if os.path.exists(filename):
            self.run_state.project_info.load()

        if self.run_state.project_info.built_commits is None:
            self.run_state.project_info.built_commits = {}


class SaveProjectInfo(icklib.BuildStep):

    def build(self):
        self.run_state.progress['step'] = 'Saving project state'
        self.run_state.logger.important('Saving project state')

        if hasattr(self.run_state, 'commits'):
            value = self.run_state.commits
            if value is not None:
                key = self.run_state.pipeline.name
                self.run_state.project_info.built_commits[key] = value

        self.run_state.project_info.save()


class CloneGits(icklib.BuildStep):

    def build(self):
        top_git_dir = self.statedir.get_git_directory(self.project.name)
        clones = []
        for git_spec in self.project.git_specs:
            dirname = os.path.normpath(
                os.path.join(top_git_dir, git_spec.subdir))
            self.run_state.progress['step'] = (
                'Cloning/updating git {}'.format(git_spec.url))
            self.run_state.logger.important(
                'git cloning {url} to {dir}',
                url=git_spec.url, dir=dirname)

            clone = self._clone_one(dirname, git_spec)
            clone.set_subdir(git_spec.subdir)
            clones.append(clone)

        self.run_state.clones = clones

    def _clone_one(self, dirname, git_spec):
        clone = icklib.GitClone()
        clone.set_dirname(dirname)
        clone.set_url(git_spec.url)
        clone.sync_from_url()
        clone.checkout(git_spec.branch)
        return clone


class FindCurrentGitCommits(icklib.BuildStep):

    def build(self):
        self.run_state.commits = {}
        for c in self.run_state.clones:
            key = '{}@{}'.format(c.url, c.subdir)
            self.run_state.commits[key] = c.get_HEAD_commit()


class FindNewCommitToBuildForShell(icklib.BuildStep):

    def build(self):
        if self._needs_to_be_built():
            self.run_state.build_using_shell = True
        else:
            self.run_state.build_using_shell = False
            self.run_state.logger.important(
                'No need to build {project_name}',
                project_name=self.project.name)

    def _needs_to_be_built(self):
        built_commits = self.run_state.project_info.built_commits
        built = built_commits.get(self.run_state.pipeline.name, None)
        return built != self.run_state.commits


class FindNewCommitToBuildForLocalShell(icklib.BuildStep):

    def build(self):
        if self._needs_to_be_built():
            self.run_state.build_using_local_shell = True
        else:
            self.run_state.build_using_local_shell = False
            self.run_state.logger.important(
                'No need to build {project_name}',
                project_name=self.project.name)

    def _needs_to_be_built(self):
        built_commits = self.run_state.project_info.built_commits
        built = built_commits.get(self.run_state.pipeline.name, None)
        return built != self.run_state.commits


class FindNewCommitToBuildForDebianCI(icklib.BuildStep):

    def build(self):
        if self._needs_to_be_built():
            self.run_state.build_using_debian_ci = True
            self.run_state.collect_debian_info = True
        else:
            self.run_state.build_using_debian_ci = False
            self.run_state.collect_debian_info = False
            self.run_state.logger.important(
                'No need to build {project_name}',
                project_name=self.project.name)

    def _needs_to_be_built(self):
        built_commits = self.run_state.project_info.built_commits
        built = built_commits.get(self.run_state.pipeline.name, None)
        return built != self.run_state.commits


class FindTagsToBuildForDebianRelease(icklib.BuildStep):

    def build(self):
        tags = self._find_tags_in_repo(
            self.project.name, self.run_state.clones[0])
        seen_tags = self.run_state.project_info.seen_tags

        if seen_tags is not None:
            not_seen_tags = [x for x in tags if x not in seen_tags]
            self.run_state.build_tags_using_debian_release = not_seen_tags
            if self.run_state.build_tags_using_debian_release:
                self.run_state.logger.important(
                    'Need to build the following release tags:')
                for tag in self.run_state.build_tags_using_debian_release:
                    self.run_state.logger.important('  {tag}', tag=tag)
            else:
                self.run_state.logger.important('No new release tags')
        else:
            self.run_state.logger.important(
                'Not seen this project before, so not building release')
            self.run_state.logger.important(
                'Remembering existing release tags for future runs')
            self.run_state.build_tags_using_debian_release = []

        self.run_state.collect_debian_info = (
            self.run_state.build_tags_using_debian_release != [])
        self.run_state.project_info.seen_tags = tags

    def _find_tags_in_repo(self, project_name, clone):
        output = cliapp.runcmd(['git', 'tag', '-l'], cwd=clone.dirname)
        return [
            tag for tag in output.splitlines()
            if self._is_release_tag(project_name, tag)]

    def _is_release_tag(self, project_name, tag):
        prefix = project_name + '-'
        if not tag.startswith(prefix):
            return False
        pattern = r'^\d+\.\d+(\.\d+)*$'
        return re.match(pattern, tag[len(prefix):])


class CreateBuildInfo(icklib.BuildStep):

    def build(self):
        # Only store a build info if we're doing a build.
        if not self.going_to_build(self.run_state):
            return

        build_infos = self.statedir.get_build_infos(self.project.name)

        build_number = self.get_next_build_number(self.run_state)
        self.run_state.project_info.save()  # Don't want to lose this if crash.
        self.run_state.build_info = self.statedir.create_new_build(
            self.project.name, build_number)
        self.run_state.build_info.status = 'RUNNING'

        self.pivot_to_real_build_log_file(self.run_state)

        # Fill in some fields from previous build, if any.
        if build_infos:
            for field in ['commit']:
                if hasattr(build_infos[-1], field):
                    value = getattr(build_infos[-1], field)
                    setattr(self.run_state.build_info, field, value)

        # Copy in new values from run_state.
        for field in ['commits']:
            if hasattr(self.run_state, field):
                value = getattr(self.run_state, field)
                if value is not None:
                    setattr(self.run_state.build_info, field, value)

        self.run_state.build_info.pipeline = self.pipeline.name

        self.run_state.build_info.save()

    def going_to_build(self, run_state):
        names = [
            'build_using_shell',
            'build_using_local_shell',
            'build_using_debian_ci',
            'build_tags_using_debian_release',
        ]
        return any(getattr(run_state, x, False) for x in names)

    def get_next_build_number(self, run_state):
        pi = run_state.project_info
        if pi.next_build_number is None:
            pi.next_build_number = 1
        result = pi.next_build_number
        pi.next_build_number += 1
        return result

    def pivot_to_real_build_log_file(self, run_state):
        filename = run_state.build_info.get_build_log_filename()
        fileurl = urlparse.urljoin('file://', filename)
        run_state.logger.important('Build log at {location}', location=fileurl)
        f = open(filename, 'w')
        f.write(run_state.log_catcher.getvalue())
        run_state.logger.drop_output_file(run_state.log_catcher)
        run_state.logger.add_output_file(f, False)


class FinishBuildInfo(icklib.BuildStep):

    def build(self):
        if self.run_state.build_info is not None:
            self.run_state.build_info.status = 'SUCCESS'
            self.run_state.build_info.duration = (
                time.time() - self.run_state.started)
            self.run_state.build_info.save()


class RunShellCommandsOnEachTarget(icklib.BuildStep):

    def build(self):
        if self.run_state.build_using_shell:
            for target in self.targets:
                self._build_on_target(self.project, target, self.run_state)

    def _build_on_target(self, project, target, run_state):
        run_state.progress['target'] = target.name
        run_state.logger.important(
            'Building {project_name} on {target_name}',
            project_name=project.name, target_name=target.name)
        with run_state.logger:
            remote_git_dir = target.mkdtemp()
            target.sync_to_target(
                run_state.clones[0].dirname, remote_git_dir, '.')
            try:
                for command in project.commands:
                    project.run_argv_on_target(
                        target, ['sh', '-euxc', command], command,
                        remote_git_dir)
            except cliapp.AppException:
                target.remove_all([remote_git_dir])
                raise
            target.remove_all([remote_git_dir])
        run_state.progress['target'] = ''


class RunLocalShellCommands(icklib.BuildStep):

    def build(self):
        if self.run_state.build_using_local_shell:
            self.build_locally(self.project, self.run_state)

    def build_locally(self, project, run_state):
        run_state.progress['target'] = 'local'
        run_state.logger.important(
            'Building {project_name} locally',
            project_name=project.name)
        git_dir = run_state.clones[0].dirname
        with run_state.logger:
            for command in project.local_commands:
                project.run_locally(
                    ['sh', '-euxc', command], command, git_dir)
        run_state.progress['target'] = ''


class CollectDebianInfoAboutTargets(icklib.BuildStep):

    def build(self):
        wanted = ['debian_arch', 'debian_codename', 'debian_release']
        if getattr(self.run_state, 'collect_debian_info', False):
            for target in self.targets:
                if not all(getattr(target, x, None) for x in wanted):
                    target.debian_arch = self._get_debian_arch(target)
                    os_release = self._get_os_release(target)
                    self._collect_from_os_release(target, os_release)

    def _get_debian_arch(self, target):
        logging.info('Finding Debian arch for %s', target.name)
        output = target.ssh_runcmd(['dpkg', '--print-architecture'])
        return output.strip()

    def _get_os_release(self, target):
        logging.info('Finding os-release for %s', target.name)
        pbuilder_argv = [
            'sudo',
            'pbuilder',
            '--execute',
            '--basetgz', target.pbuilder_ci_tgz,
            '--',
            '/bin/cat', '/etc/os-release',
        ]
        output = target.ssh_runcmd(pbuilder_argv)
        return icklib.parse_os_release(output)

    def _collect_from_os_release(self, target, os_release):
        if os_release.debian_codename:
            codename = os_release.debian_codename
        elif os_release.pretty_codename:
            codename = os_release.pretty_codename
        else:
            assert False, 'os-release has no useful Debian codename'

        # We treat 'sid' specially.
        if codename == 'sid':
            codename = 'unstable'

        target.debian_codename = codename

        if codename == 'unstable':
            release = 'unstable'
        elif os_release.debian_release:
            release = 'debian' + os_release.debian_release
        else:
            assert False, 'os-release has no useful Debian release'
        target.debian_release = release


class CreateCITarball(icklib.BuildStep):

    def build(self):
        if not self.run_state.build_using_debian_ci:
            return

        self.run_state.progress['step'] = 'Creating CI tarball'
        self.run_state.logger.important('Creating CI tarball')
        with self.run_state.logger:
            tarball_target = self.targets[0]
            remote_tmp = tarball_target.mkdtemp()

            remote_git_dir = tarball_target.sync_to_target(
                self.run_state.clones[0].dirname, remote_tmp, 'git')
            remote_artifacts_dir = tarball_target.sync_to_target(
                self.run_state.build_info.create_artifacts_directory(),
                remote_tmp, 'artifacts')

            cleanly_path = self.project.copy_cleanly_to_target(
                tarball_target, remote_tmp)
            self.project.run_cleanly(
                tarball_target, remote_git_dir, cleanly_path,
                ['--results', remote_artifacts_dir,
                 '--ci',
                 '--buildno', str(self.run_state.build_info.build_number),
                 '--debian-release', tarball_target.debian_release,
                 '--debian-codename', tarball_target.debian_codename,
                 'tarball'])

            tarball_target.sync_from_target(
                remote_artifacts_dir,
                self.run_state.build_info.create_artifacts_directory())

            tarball_target.remove_all([remote_tmp])


class CreateReleaseTarballs(icklib.BuildStep):

    def build(self):
        target = self.targets[0]
        for tag in sorted(self.run_state.build_tags_using_debian_release):
            self.run_state.progress['step'] = (
                'Create release tarball for {}'.format(tag))
            self.run_state.logger.important(
                'Creating release tarball for tag {tag}',
                tag=tag)
            with self.run_state.logger:
                remote_tmp = target.mkdtemp()
                self.run_state.clones[0].checkout(tag)
                remote_git_dir = target.sync_to_target(
                    self.run_state.clones[0].dirname, remote_tmp, 'git')
                remote_artifacts_dir = target.sync_to_target(
                    self.run_state.build_info.create_artifacts_directory(),
                    remote_tmp, 'artifacts')

                cleanly_path = self.project.copy_cleanly_to_target(
                    target, remote_tmp)
                self.project.run_cleanly(
                    target, remote_git_dir, cleanly_path,
                    ['--results', remote_artifacts_dir,
                     '--release',
                     '--buildno', str(self.run_state.build_info.build_number),
                     '--debian-release', target.debian_release,
                     '--debian-codename', target.debian_codename,
                     'tarball'])

                target.sync_from_target(
                    remote_artifacts_dir,
                    self.run_state.build_info.create_artifacts_directory())

                target.remove_all([remote_tmp])


class CreateDebianSourcePackagesForCI(icklib.BuildStep):

    def build(self):
        if not self.run_state.build_using_debian_ci:
            return

        dsc_targets = []
        codenames = set()
        for target in self.targets:
            if target.debian_codename not in codenames:
                dsc_targets.append(target)
                codenames.add(target.debian_codename)

        for target in dsc_targets:
            self.run_state.progress['step'] = (
                'Create dsc for {}'.format(target.debian_codename))
            self.run_state.logger.important(
                'Creating dsc for {codename}',
                codename=target.debian_codename)
            with self.run_state.logger:
                remote_tmp = target.mkdtemp()
                remote_git_dir = target.sync_to_target(
                    self.run_state.clones[0].dirname, remote_tmp, 'git')
                remote_artifacts_dir = target.sync_to_target(
                    self.run_state.build_info.create_artifacts_directory(),
                    remote_tmp, 'artifacts')

                cleanly_path = self.project.copy_cleanly_to_target(
                    target, remote_tmp)
                self.project.run_cleanly(
                    target, remote_git_dir, cleanly_path,
                    ['--results', remote_artifacts_dir,
                     '--ci',
                     '--buildno', str(self.run_state.build_info.build_number),
                     '--debian-release', target.debian_release,
                     '--debian-codename', target.debian_codename,
                     'dsc'])

                target.sync_from_target(
                    remote_artifacts_dir,
                    self.run_state.build_info.create_artifacts_directory())

                target.remove_all([remote_tmp])


class CreateDebianSourcePackagesForRelease(icklib.BuildStep):

    def build(self):
        if not self.run_state.build_tags_using_debian_release:
            return

        dsc_targets = self._get_dsc_targets(self.targets)
        for tag in self.run_state.build_tags_using_debian_release:
            for target in dsc_targets:
                self.run_state.progress['step'] = (
                    'Create dsc for {} for tag {}'.format(
                        target.debian_codename, tag))
                self.run_state.logger.important(
                    'Creating dsc for {codename} for tag {tag}',
                    codename=target.debian_codename,
                    tag=tag)
                with self.run_state.logger:
                    remote_tmp = target.mkdtemp()
                    self.run_state.clones[0].checkout(tag)
                    remote_git_dir = target.sync_to_target(
                        self.run_state.clones[0].dirname, remote_tmp, 'git')
                    remote_artifacts_dir = target.sync_to_target(
                        self.run_state.build_info.create_artifacts_directory(),
                        remote_tmp, 'artifacts')

                    cleanly_path = self.project.copy_cleanly_to_target(
                        target, remote_tmp)
                    self.project.run_cleanly(
                        target, remote_git_dir, cleanly_path,
                        ['--results', remote_artifacts_dir,
                         '--release',
                         '--buildno',
                         str(self.run_state.build_info.build_number),
                         '--debian-release', target.debian_release,
                         '--debian-codename', target.debian_codename,
                         'dsc'])

                    target.sync_from_target(
                        remote_artifacts_dir,
                        self.run_state.build_info.create_artifacts_directory())

                    target.remove_all([remote_tmp])

    def _get_dsc_targets(self, targets):
        dsc_targets = []
        codenames = set()
        for target in targets:
            if target.debian_codename not in codenames:
                dsc_targets.append(target)
                codenames.add(target.debian_codename)
        return dsc_targets


class CreateCIDebianBinaryPackages(icklib.BuildStep):

    def build(self):
        if not self.run_state.build_using_debian_ci:
            return

        self.run_state.progress['step'] = 'Creating debs'
        self.run_state.logger.important('Creating debs')
        with self.run_state.logger:
            archs = self._get_architectures_from_source(
                self.run_state.clones[0])

            arch_all_built = set()
            for target in self.targets:
                remote_tmp = target.mkdtemp()
                remote_git_dir = target.sync_to_target(
                    self.run_state.clones[0].dirname, remote_tmp, 'git')
                remote_artifacts_dir = target.sync_to_target(
                    self.run_state.build_info.create_artifacts_directory(),
                    remote_tmp, 'artifacts')

                cleanly_path = self.project.copy_cleanly_to_target(
                    target, remote_tmp)
                cleanly_options = [
                    '--results', remote_artifacts_dir,
                    '--pbuilder-tgz', target.pbuilder_ci_tgz,
                    '--ci',
                    '--buildno', str(self.run_state.build_info.build_number),
                    '--debian-release', target.debian_release,
                    '--debian-codename', target.debian_codename,
                ]

                build_it = True
                if 'all' in archs:
                    if target.debian_codename not in arch_all_built:
                        cleanly_options += ['--build-arch-all-also']
                        arch_all_built.add(target.debian_codename)
                    elif archs == ['all']:
                        build_it = False

                if build_it:
                    self.project.run_cleanly(
                        target, remote_git_dir, cleanly_path,
                        cleanly_options + ['deb'])

                    target.sync_from_target(
                        remote_artifacts_dir,
                        self.run_state.build_info.create_artifacts_directory())

                target.remove_all([remote_tmp])

    def _get_architectures_from_source(self, clone):
        archs = set()
        control = os.path.join(clone.dirname, 'debian', 'control')
        with open(control) as f:
            for line in f:
                if line.startswith('Architecture: '):
                    for arch in line.split()[1:]:
                        archs.add(arch)
        return list(archs)


class CreateDebianBinaryPackagesForRelease(icklib.BuildStep):

    def build(self):
        for tag in self.run_state.build_tags_using_debian_release:
            self.run_state.progress['what'] = (
                'Building debs for release {}'.format(tag))
            self.run_state.logger.important(
                'Building debs for release {tag}',
                tag=tag)
            with self.run_state.logger:
                self.run_state.clones[0].checkout(tag)

                archs = self._get_architectures_from_source(
                    self.run_state.clones[0])

                arch_all_built = set()
                for target in self.targets:
                    remote_tmp = target.mkdtemp()
                    remote_git_dir = target.sync_to_target(
                        self.run_state.clones[0].dirname, remote_tmp, 'git')
                    remote_artifacts_dir = target.sync_to_target(
                        self.run_state.build_info.create_artifacts_directory(),
                        remote_tmp, 'artifacts')

                    cleanly_path = self.project.copy_cleanly_to_target(
                        target, remote_tmp)
                    cleanly_options = [
                        '--results', remote_artifacts_dir,
                        '--pbuilder-tgz', target.pbuilder_ci_tgz,
                        '--release',
                        '--buildno',
                        str(self.run_state.build_info.build_number),
                        '--debian-release', target.debian_release,
                        '--debian-codename', target.debian_codename,
                    ]

                    build_it = True
                    if 'all' in archs:
                        if target.debian_codename not in arch_all_built:
                            cleanly_options += ['--build-arch-all-also']
                            arch_all_built.add(target.debian_codename)
                        elif archs == ['all']:
                            build_it = False

                    if build_it:
                        self.project.run_cleanly(
                            target, remote_git_dir, cleanly_path,
                            cleanly_options + ['deb'])

                        build_info = self.run_state.build_info
                        target.sync_from_target(
                            remote_artifacts_dir,
                            build_info.create_artifacts_directory())

                    target.remove_all([remote_tmp])

    def _get_architectures_from_source(self, clone):
        archs = set()
        control = os.path.join(clone.dirname, 'debian', 'control')
        with open(control) as f:
            for line in f:
                if line.startswith('Architecture: '):
                    for arch in line.split()[1:]:
                        archs.add(arch)
        return list(archs)


class FindDebianChangesFiles(icklib.BuildStep):

    def build(self):
        self.run_state.changes_files = []
        if self.building():
            artifacts = self.run_state.build_info.create_artifacts_directory()
            for changes in glob.glob(os.path.join(artifacts, '*.changes')):
                changes_url = urlparse.urljoin(
                    'file://',
                    urllib.quote(os.path.abspath(changes)))
                self.run_state.changes_files.append(changes_url)

    def building(self):
        names = [
            'build_using_shell',
            'build_using_debian_ci',
            'build_tags_using_debian_release',
        ]
        return any(getattr(self.run_state, x, False) for x in names)


class SetupAPTRepository(icklib.BuildStep):

    def build(self):
        if not self.run_state.changes_files:
            return

        self.run_state.progress['step'] = 'Prepare APT repository'
        self.run_state.logger.important('Prepare APT repository')
        repopath = self.statedir.get_apt_directory()
        confpath = os.path.join(repopath, 'conf')
        if not os.path.exists(confpath):
            os.makedirs(confpath)

        self._refresh_file(
            os.path.join(confpath, 'distributions'),
            self._make_reprepro_distributions(self.targets))

        self._refresh_file(
            os.path.join(confpath, 'incoming'),
            self._make_reprepro_incoming(self.targets))

        for base in ('incoming', 'incoming.temp'):
            path = os.path.join(repopath, base)
            if not os.path.exists(path):
                os.makedirs(path)

    def _make_reprepro_distributions(self, targets):
        archs = set(t.debian_arch for t in targets)
        codenames = set(t.debian_codename for t in targets)

        if self.project.repo_signing_key is None:
            signwith = ''
        else:
            signwith = 'SignWith: {}\n'.format(self.project.repo_signing_key)

        result = ''
        for codename in codenames:
            ci_stanza = (
                'Codename: {codename}-ci\n'
                'Suite: {codename}\n'
                'Origin: ick\n'
                'Description: packages built by ick for CI\n'
                'Architectures: source {archlist}\n'
                'Components: main\n'
                '{signwith}'
                '\n'
            )
            result += ci_stanza.format(
                codename=codename,
                archlist=' '.join(archs),
                signwith=signwith)

            release_stanza = (
                'Codename: {codename}\n'
                'Suite: {codename}\n'
                'Origin: ick\n'
                'Description: packages built by ick for release\n'
                'Architectures: source {archlist}\n'
                'Components: main\n'
                '{signwith}'
                '\n'
            )
            result += release_stanza.format(
                codename=codename,
                archlist=' '.join(archs),
                signwith=signwith)

        return result

    def _make_reprepro_incoming(self, targets):
        release_codenames = [t.debian_codename for t in targets]
        ci_codenames = [x + '-ci' for x in release_codenames]
        codenames = ' '.join(release_codenames + ci_codenames)

        stanza = (
            'Name: default\n'
            'IncomingDir: incoming\n'
            'TempDir: incoming.temp\n'
            'Cleanup: on_error\n'
            'Allow: {codenames}\n')
        return stanza.format(codenames=codenames)

    def _refresh_file(self, pathname, contents):
        if os.path.exists(pathname):
            with open(pathname) as f:
                old_contents = f.read()
        else:
            old_contents = None
        if old_contents != contents:
            with open(pathname, 'w') as f:
                f.write(contents)


class UploadDebianPackagesToCIRepo(icklib.BuildStep):

    def build(self):
        if not self.run_state.changes_files:
            return

        self.run_state.progress['step'] = (
            'Upload built packages to APT repository')
        self.run_state.logger.important(
            'Upload built packages to APT repository')
        with self.run_state.logger:
            repopath = self.statedir.get_apt_directory()
            incoming = os.path.join(repopath, 'incoming')
            line_logger = LineLogger(
                self.run_state.progress, self.run_state.logger)
            for changes_url in self.run_state.changes_files:
                # dget doesn't URL-decode file:// URLs. So we do that.
                changes_url = urllib.unquote(changes_url)
                cliapp.runcmd(
                    ['dget', '-du', changes_url],
                    cwd=incoming,
                    stderr=subprocess.STDOUT,
                    stdout_callback=line_logger.log_lines)

            exit_code, stdout, stderr = cliapp.runcmd_unchecked(
                ['reprepro', 'processincoming', 'default'],
                cwd=repopath)
            line_logger.log_lines(stdout)
            line_logger.log_lines(stderr)
            if exit_code != 0:
                pattern = r"Name '.*-dbgsym' of binary '.*' is not listed"
                if re.search(pattern, stderr):
                    self.run_state.logger.chatty(
                        'Ignoring reprepro problem wrt -dbgsym')
                    cliapp.runcmd(
                        ['reprepro', 'export'],
                        cwd=repopath,
                        stdout_callback=line_logger.log_lines,
                        stderr=subprocess.STDOUT)
                else:
                    sys.exit(exit_code)


class PublishDebianPackages(icklib.BuildStep):

    def build(self):
        if not self.run_state.changes_files or not self.project.dput_targets:
            return

        self.run_state.progress['step'] = (
            'Publish built release packages to public APT repositories')
        self.run_state.logger.important(
            'Publish built release packages to public APT repositories')
        with self.run_state.logger:
            tempdir = self._copy_packages()
            self._remove_dbgsym_packages_from_changes_files(tempdir)
            self._debsign(tempdir)
            self._dput(tempdir)
            shutil.rmtree(tempdir)

    def _copy_packages(self):
        tempdir = tempfile.mkdtemp()
        artifacts_dir = self.run_state.build_info.get_artifacts_directory()
        self.project.run_locally(
            ['rsync', '-a', artifacts_dir + '/.', tempdir + '/.'],
            'rsync artifacts', '.')
        return tempdir

    def _remove_dbgsym_packages_from_changes_files(self, tempdir):
        for filename in self._changes_files(tempdir):
            with open(filename) as f:
                text = f.read()
            modified = ''.join(
                '{}\n'.format(line)
                for line in text.splitlines()
                if '-dbgsym' not in line)
            with open(filename, 'w') as f:
                f.write(modified)

    def _changes_files(self, tempdir):
        return glob.glob(os.path.join(tempdir, '*.changes'))

    def _debsign(self, tempdir):
        files = self._changes_files(tempdir)
        self.project.run_locally(
            ['debsign'] + files, 'debsign *.changes', tempdir)

    def _dput(self, tempdir):
        files = self._changes_files(tempdir)
        for dput_target in self.project.dput_targets:
            self.project.run_locally(
                ['dput', dput_target] + files,
                'dput {} *.changes'.format(dput_target),
                tempdir)


class LineLogger(object):  # pragma: no cover

    def __init__(self, progress, logger):
        self._progress = progress
        self._progress_lines = []
        self._logger = logger
        self._buffer = ''
        self._lineno = 0

    def log_lines(self, text):
        self._buffer += text
        while '\n' in self._buffer:
            line, self._buffer = self._buffer.split('\n', 1)
            self._log_line(line)

        # Return empty string so that cliapp doesn't collect
        # potentially large amounts of text pointlessly.
        return ''

    def _log_line(self, line):
        self._lineno += 1
        line = '{lineno:>04}: {line}'.format(lineno=self._lineno, line=line)
        logging.debug('STDOUT: %s', line)
        self._logger.chatty('{line}', line=line)

        lines = self._progress_lines
        max_lines = self._progress.max_output_lines

        lines.append(line)
        if len(lines) > max_lines:
            del lines[:-max_lines]
        for i, line in enumerate(lines):
            self._progress['line%d' % i] = line
        for i in range(len(lines), max_lines + 1):
            self._progress['line%d' % i] = ''

    def log_rest(self):
        if self._buffer:
            self._log_line(self._buffer)


def create_projects_from_ick(ick, wanted_names):
    pipelines = {
        'shell': [
            LoadProjectInfo,
            CloneGits,
            FindCurrentGitCommits,
            FindNewCommitToBuildForShell,
            CreateBuildInfo,
            RunShellCommandsOnEachTarget,
            FinishBuildInfo,
            SaveProjectInfo,
        ],
        'local-shell': [
            LoadProjectInfo,
            CloneGits,
            FindCurrentGitCommits,
            FindNewCommitToBuildForLocalShell,
            CreateBuildInfo,
            RunLocalShellCommands,
            FinishBuildInfo,
            SaveProjectInfo,
        ],
        'debian-ci': [
            LoadProjectInfo,
            CloneGits,
            FindCurrentGitCommits,
            FindNewCommitToBuildForDebianCI,
            CreateBuildInfo,
            CollectDebianInfoAboutTargets,
            CreateCITarball,
            CreateDebianSourcePackagesForCI,
            CreateCIDebianBinaryPackages,
            FindDebianChangesFiles,
            SetupAPTRepository,
            UploadDebianPackagesToCIRepo,
            FinishBuildInfo,
            SaveProjectInfo,
        ],
        'debian-release': [
            LoadProjectInfo,
            CloneGits,
            FindCurrentGitCommits,
            FindTagsToBuildForDebianRelease,
            CreateBuildInfo,
            CollectDebianInfoAboutTargets,
            CreateReleaseTarballs,
            CreateDebianSourcePackagesForRelease,
            CreateDebianBinaryPackagesForRelease,
            FindDebianChangesFiles,
            SetupAPTRepository,
            UploadDebianPackagesToCIRepo,
            PublishDebianPackages,
            FinishBuildInfo,
            SaveProjectInfo,
        ],
    }

    projects = []
    if not wanted_names:
        wanted_names = ick.get('projects', {}).keys()

    for project_name, project_dict in ick.get('projects', {}).items():
        if project_name in wanted_names:
            p = Project(project_name)
            for git_spec in create_git_specs_for_project(project_dict):
                p.add_git_spec(git_spec)
            if 'repo-signing-key' in ick:
                p.set_repo_signing_key(ick['repo-signing-key'])
            if 'shell' in project_dict:
                p.add_shell_command(project_dict['shell'])
            if 'local-shell' in project_dict:
                p.add_local_command(project_dict['local-shell'])
            if 'debian-release' in project_dict:
                release = project_dict['debian-release']
                dput_targets = release.get('dput-targets', [])
                for dput_target in dput_targets:
                    p.add_dput_target(dput_target)
            p.set_build_pipelines(
                dict((x, pipelines[x]) for x in project_dict['pipelines']))
            p.set_env(project_dict.get('env', {}))

            projects.append(p)

    return projects


def create_git_specs_for_project(project_dict):
    if 'git' in project_dict:
        yield GitSpec(project_dict['git'], project_dict['branch'], '.')
    elif 'gits' in project_dict:
        for git_dict in project_dict['gits']:
            yield GitSpec(
                git_dict['git'], git_dict['branch'], git_dict['root'])
